@@ -1,5 +1,6 @@
 /* GtkToolkit.java -- Implements an AWT Toolkit using GTK for peers
-   Copyright (C) 1998, 1999, 2002, 2003, 2004, 2005  Free Software Foundation, Inc.
+   Copyright (C) 1998, 1999, 2002, 2003, 2004, 2005, 2006, 2007
+   Free Software Foundation, Inc.
 
 This file is part of GNU Classpath.
 
@@ -39,31 +40,87 @@ exception statement from your version. */
 package gnu.java.awt.peer.gtk;
 
 import gnu.classpath.Configuration;
+
+import gnu.java.awt.AWTUtilities;
 import gnu.java.awt.EmbeddedWindow;
+import gnu.java.awt.dnd.GtkMouseDragGestureRecognizer;
+import gnu.java.awt.dnd.peer.gtk.GtkDragSourceContextPeer;
 import gnu.java.awt.peer.ClasspathFontPeer;
-import gnu.java.awt.peer.ClasspathTextLayoutPeer;
 import gnu.java.awt.peer.EmbeddedWindowPeer;
 
-import java.awt.*;
+import java.awt.AWTException;
+import java.awt.Button;
+import java.awt.Canvas;
+import java.awt.Checkbox;
+import java.awt.CheckboxMenuItem;
+import java.awt.Choice;
+import java.awt.Component;
+import java.awt.Cursor;
+import java.awt.Dialog;
+import java.awt.Dimension;
+import java.awt.EventQueue;
+import java.awt.FileDialog;
+import java.awt.Font;
+import java.awt.FontMetrics;
+import java.awt.Frame;
+import java.awt.GraphicsDevice;
+import java.awt.GraphicsEnvironment;
+import java.awt.HeadlessException;
+import java.awt.Image;
+import java.awt.Label;
+import java.awt.List;
+import java.awt.Menu;
+import java.awt.MenuBar;
+import java.awt.MenuItem;
+import java.awt.Panel;
+import java.awt.Point;
+import java.awt.PopupMenu;
+import java.awt.PrintJob;
+import java.awt.Rectangle;
+import java.awt.ScrollPane;
+import java.awt.Scrollbar;
+import java.awt.TextArea;
+import java.awt.TextField;
+import java.awt.Window;
 import java.awt.datatransfer.Clipboard;
 import java.awt.dnd.DragGestureEvent;
+import java.awt.dnd.DragGestureListener;
+import java.awt.dnd.DragGestureRecognizer;
+import java.awt.dnd.DragSource;
+import java.awt.dnd.InvalidDnDOperationException;
 import java.awt.dnd.peer.DragSourceContextPeer;
-import java.awt.font.FontRenderContext;
+import java.awt.font.TextAttribute;
 import java.awt.im.InputMethodHighlight;
-import java.awt.image.BufferedImage;
 import java.awt.image.ColorModel;
 import java.awt.image.DirectColorModel;
-import java.awt.image.ImageConsumer;
 import java.awt.image.ImageObserver;
 import java.awt.image.ImageProducer;
-import java.awt.peer.*;
+import java.awt.peer.ButtonPeer;
+import java.awt.peer.CanvasPeer;
+import java.awt.peer.CheckboxMenuItemPeer;
+import java.awt.peer.CheckboxPeer;
+import java.awt.peer.ChoicePeer;
+import java.awt.peer.DialogPeer;
+import java.awt.peer.FileDialogPeer;
+import java.awt.peer.FontPeer;
+import java.awt.peer.FramePeer;
+import java.awt.peer.LabelPeer;
+import java.awt.peer.ListPeer;
+import java.awt.peer.MenuBarPeer;
+import java.awt.peer.MenuItemPeer;
+import java.awt.peer.MenuPeer;
+import java.awt.peer.MouseInfoPeer;
+import java.awt.peer.PanelPeer;
+import java.awt.peer.PopupMenuPeer;
+import java.awt.peer.RobotPeer;
+import java.awt.peer.ScrollPanePeer;
+import java.awt.peer.ScrollbarPeer;
+import java.awt.peer.TextAreaPeer;
+import java.awt.peer.TextFieldPeer;
+import java.awt.peer.WindowPeer;
 import java.io.InputStream;
 import java.net.URL;
-import java.text.AttributedString;
 import java.util.HashMap;
-import java.util.HashSet;
-import java.util.Hashtable;
-import java.util.Iterator;
 import java.util.LinkedHashMap;
 import java.util.Map;
 import java.util.Properties;
@@ -77,42 +134,58 @@ import javax.imageio.spi.IIORegistry;
    this class.  If getPeer() ever goes away, we can implement a hash table
    that will keep up with every window's peer, but for now this is faster. */
 
-/**
- * This class accesses a system property called
- * <tt>gnu.java.awt.peer.gtk.Graphics</tt>.  If the property is defined and
- * equal to "Graphics2D", the cairo-based GdkGraphics2D will be used in
- * drawing contexts. Any other value will cause the older GdkGraphics
- * object to be used.
- */
 public class GtkToolkit extends gnu.java.awt.ClasspathToolkit
 {
-  Hashtable containers = new Hashtable();
-  static EventQueue q;
-  static boolean useGraphics2dSet;
-  static boolean useGraphics2d;
-  static Thread mainThread;
+  static final Object GTK_LOCK;
 
-  public static boolean useGraphics2D()
+  private static EventQueue q;
+
+  static native void gtkInit(int portableNativeSync, Object lock);
+
+  static native void gtkMain();
+
+  static native void gtkQuit();
+
+  /**
+   * Initializes field IDs that are used by native code.
+   */
+  private static native void initIDs();
+
+  /**
+   * True when the field IDs are already initialized, false otherwise.
+   */
+  private static boolean initializedGlobalIDs = false;
+
+  /**
+   * Initializes some global fieldIDs for use in the native code. This is
+   * called by a couple of classes in the GTK peers to ensure that
+   * some necessary stuff is loaded.
+   */
+  static synchronized void initializeGlobalIDs()
   {
-    if (useGraphics2dSet)
-      return useGraphics2d;
-    useGraphics2d = System.getProperty("gnu.java.awt.peer.gtk.Graphics", 
-                                       "Graphics").equals("Graphics2D");
-    useGraphics2dSet = true;
-    return useGraphics2d;
+    if (! initializedGlobalIDs)
+      {
+        initIDs();
+        initializedGlobalIDs = true;
+      }
   }
 
-  static native void gtkInit(int portableNativeSync);
-
   static
   {
-    if (Configuration.INIT_LOAD_LIBRARY)
-      System.loadLibrary("gtkpeer");
+    if (true) // GCJ LOCAL
+      {
+        System.loadLibrary("gtkpeer");
+      }
+
+    /**
+     * Gotta do that first.
+     */
+    initializeGlobalIDs();
 
     int portableNativeSync;     
     String portNatSyncProp = 
       System.getProperty("gnu.classpath.awt.gtk.portable.native.sync");
-    
+      
     if (portNatSyncProp == null)
       portableNativeSync = -1;  // unset
     else if (Boolean.valueOf(portNatSyncProp).booleanValue())
@@ -120,16 +193,8 @@ public class GtkToolkit extends gnu.java.awt.ClasspathToolkit
     else
       portableNativeSync = 0;   // false
 
-    gtkInit(portableNativeSync);
-
-    mainThread = new Thread ("GTK main thread")
-      {
-        public void run ()
-        {
-          gtkMain ();
-        }
-      };
-    mainThread.start ();
+    GTK_LOCK = new String("GTK LOCK");
+    gtkInit(portableNativeSync, GTK_LOCK);
   }
 
   public GtkToolkit ()
@@ -137,17 +202,21 @@ public class GtkToolkit extends gnu.java.awt.ClasspathToolkit
   }
 
   public native void beep();
+
   private native void getScreenSizeDimensions(int[] xy);
   
   public int checkImage (Image image, int width, int height, 
-			 ImageObserver observer) 
+                         ImageObserver observer) 
   {
     int status = ImageObserver.ALLBITS 
       | ImageObserver.WIDTH 
       | ImageObserver.HEIGHT;
 
     if (image instanceof GtkImage)
-	return ((GtkImage) image).checkImage (observer);
+      return ((GtkImage) image).checkImage (observer);
+
+    if (image instanceof AsyncImage)
+      return ((AsyncImage) image).checkImage(observer);
 
     if (observer != null)
       observer.imageUpdate (image, status,
@@ -159,137 +228,71 @@ public class GtkToolkit extends gnu.java.awt.ClasspathToolkit
   }
 
   /** 
-   * A helper class to return to clients in cases where a BufferedImage is
-   * desired but its construction fails.
+   * Helper to return either a Image -- the argument -- or a
+   * GtkImage with the errorLoading flag set if the argument is null.
    */
-  private class GtkErrorImage extends Image
-  {
-    public GtkErrorImage()
-    {
-    }
-
-    public int getWidth(ImageObserver observer)
-    {
-      return -1;
-    }
-
-    public int getHeight(ImageObserver observer)
-    {
-      return -1;
-    }
-
-    public ImageProducer getSource()
-    {
-
-      return new ImageProducer() 
-        {          
-          HashSet consumers = new HashSet();          
-          public void addConsumer(ImageConsumer ic)
-          {
-            consumers.add(ic);
-          }
-
-          public boolean isConsumer(ImageConsumer ic)
-          {
-            return consumers.contains(ic);
-          }
-
-          public void removeConsumer(ImageConsumer ic)
-          {
-            consumers.remove(ic);
-          }
-
-          public void startProduction(ImageConsumer ic)
-          {
-            consumers.add(ic);
-            Iterator i = consumers.iterator();
-            while(i.hasNext())
-              {
-                ImageConsumer c = (ImageConsumer) i.next();
-                c.imageComplete(ImageConsumer.IMAGEERROR);
-              }
-          }
-          public void requestTopDownLeftRightResend(ImageConsumer ic)
-          {
-            startProduction(ic);
-          }        
-        };
-    }
-
-    public Graphics getGraphics() 
-    { 
-      return null; 
-    }
-
-    public Object getProperty(String name, ImageObserver observer)
-    {
-      return null;
-    }
-    public Image getScaledInstance(int width, int height, int flags)
-    {
-      return new GtkErrorImage();
-    }
-
-    public void flush() 
-    {
-    }
-  }
-
-
-  /** 
-   * Helper to return either a BufferedImage -- the argument -- or a
-   * GtkErrorImage if the argument is null.
-   */
-
-  private Image bufferedImageOrError(BufferedImage b)
+  static Image imageOrError(Image b)
   {
     if (b == null) 
-      return new GtkErrorImage();
+      return GtkImage.getErrorImage();
     else
       return b;
   }
 
-
   public Image createImage (String filename)
   {
     if (filename.length() == 0)
       return new GtkImage ();
-
-    if (useGraphics2D())
-      return bufferedImageOrError(GdkPixbufDecoder.createBufferedImage (filename));
-    else
-      return new GtkImage (filename);
+    
+    Image image;
+    try
+      {
+        image = CairoSurface.getBufferedImage( new GtkImage( filename ) );
+      }
+    catch (IllegalArgumentException iae)
+      {
+        image = null;
+      }
+    return imageOrError(image);
   }
 
   public Image createImage (URL url)
   {
-    if (useGraphics2D())
-      return bufferedImageOrError(GdkPixbufDecoder.createBufferedImage (url));
-    else
-      return new GtkImage (url);
+    return new AsyncImage(url);
   }
 
   public Image createImage (ImageProducer producer) 
   {
-    if (useGraphics2D())
-      return bufferedImageOrError(GdkPixbufDecoder.createBufferedImage (producer));
-    else
-      return new GtkImage (producer);
+    if (producer == null)
+      return null;
+      
+    Image image;
+    try
+      {
+        image = CairoSurface.getBufferedImage( new GtkImage( producer ) );
+      }
+    catch (IllegalArgumentException iae)
+      {
+        image = null;
+      }
+    return imageOrError(image);
   }
 
   public Image createImage (byte[] imagedata, int imageoffset,
 			    int imagelength)
   {
-    if (useGraphics2D())
-      return bufferedImageOrError(GdkPixbufDecoder.createBufferedImage (imagedata,
-                                                   imageoffset, 
-                                                                        imagelength));
-    else
+    Image image;
+    try
       {
-        byte[] datacopy = new byte[imagelength];
-        System.arraycopy (imagedata, imageoffset, datacopy, 0, imagelength);
-        return new GtkImage (datacopy);
+        byte[] data = new byte[ imagelength ];
+        System.arraycopy(imagedata, imageoffset, data, 0, imagelength);
+        image = CairoSurface.getBufferedImage( new GtkImage( data ) );
       }
+    catch (IllegalArgumentException iae)
+      {
+        image = null;
+      }
+    return imageOrError(image);
   }
   
   /**
@@ -300,7 +303,7 @@ public class GtkToolkit extends gnu.java.awt.ClasspathToolkit
    */  
   public ImageProducer createImageProducer(URL url)
   {
-    return new GdkPixbufDecoder(url);  
+    return createImage( url ).getSource();
   }
 
   /**
@@ -311,22 +314,22 @@ public class GtkToolkit extends gnu.java.awt.ClasspathToolkit
   {
     /* Return the GDK-native ABGR format */
     return new DirectColorModel(32, 
-				0x000000FF,
-				0x0000FF00,
-				0x00FF0000,
-				0xFF000000);
+                                0x000000FF,
+                                0x0000FF00,
+                                0x00FF0000,
+                                0xFF000000);
   }
 
   public String[] getFontList () 
   {
     return (new String[] { "Dialog", 
-			   "DialogInput", 
-			   "Monospaced", 
-			   "Serif", 
-			   "SansSerif" });
+                           "DialogInput", 
+                           "Monospaced", 
+                           "Serif",
+                           "SansSerif" });
   }
 
-  private class LRUCache extends LinkedHashMap
+  static class LRUCache<K,V> extends LinkedHashMap<K,V>
   {    
     int max_entries;
     public LRUCache(int max)
@@ -340,30 +343,19 @@ public class GtkToolkit extends gnu.java.awt.ClasspathToolkit
     }
   }
 
-  private LRUCache fontCache = new LRUCache(50);
-  private LRUCache metricsCache = new LRUCache(50);
-  private LRUCache imageCache = new LRUCache(50);
+  private LRUCache<Map,ClasspathFontPeer> fontCache =
+    new LRUCache<Map,ClasspathFontPeer>(50);
+  private LRUCache<Object,Image> imageCache = new LRUCache<Object,Image>(50);
 
   public FontMetrics getFontMetrics (Font font) 
   {
-    synchronized (metricsCache)
-      {
-        if (metricsCache.containsKey(font))
-          return (FontMetrics) metricsCache.get(font);
-      }
-
-    FontMetrics m = new GdkFontMetrics (font);
-    synchronized (metricsCache)
-      {
-        metricsCache.put(font, m);
-      }
-    return m;
+    return ((GdkFontPeer) font.getPeer()).getFontMetrics(font);
   }
 
   public Image getImage (String filename) 
   {
     if (imageCache.containsKey(filename))
-      return (Image) imageCache.get(filename);
+      return imageCache.get(filename);
     else
       {
         Image im = createImage(filename);
@@ -375,7 +367,7 @@ public class GtkToolkit extends gnu.java.awt.ClasspathToolkit
   public Image getImage (URL url) 
   {
     if (imageCache.containsKey(url))
-      return (Image) imageCache.get(url);
+      return imageCache.get(url);
     else
       {
         Image im = createImage(url);
@@ -386,6 +378,11 @@ public class GtkToolkit extends gnu.java.awt.ClasspathToolkit
 
   public PrintJob getPrintJob (Frame frame, String jobtitle, Properties props) 
   {
+    SecurityManager sm;
+    sm = System.getSecurityManager();
+    if (sm != null)
+      sm.checkPrintJobAccess();
+
     return null;
   }
 
@@ -404,7 +401,16 @@ public class GtkToolkit extends gnu.java.awt.ClasspathToolkit
     if (secman != null)
       secman.checkSystemClipboardAccess();
 
-    return GtkClipboard.getInstance();
+    return GtkClipboard.getClipboardInstance();
+  }
+
+  public Clipboard getSystemSelection()
+  {
+    SecurityManager secman = System.getSecurityManager();
+    if (secman != null)
+      secman.checkSystemClipboardAccess();
+    
+    return GtkClipboard.getSelectionInstance();
   }
 
   /**
@@ -412,12 +418,19 @@ public class GtkToolkit extends gnu.java.awt.ClasspathToolkit
    * assumes the image is already prepared for rendering.
    */
   public boolean prepareImage (Image image, int width, int height, 
-			       ImageObserver observer) 
+                               ImageObserver observer) 
   {
     /* GtkImages are always prepared, as long as they're loaded. */
     if (image instanceof GtkImage)
-      return ((((GtkImage)image).checkImage (observer) & 
-	       ImageObserver.ALLBITS) != 0);
+      return ((((GtkImage)image).checkImage (observer)
+               & ImageObserver.ALLBITS) != 0);
+
+    if (image instanceof AsyncImage)
+      {
+        AsyncImage aImg = (AsyncImage) image;
+        aImg.addObserver(observer);
+        return aImg.realImage != null;
+      }
 
     /* Assume anything else is too */
     return true;
@@ -438,116 +451,141 @@ public class GtkToolkit extends gnu.java.awt.ClasspathToolkit
     /* Make the Peer reflect the state of the Component */
     if (! (c instanceof Window))
       {
-	cp.setCursor (c.getCursor ());
+        cp.setCursor (c.getCursor ());
 	
-	Rectangle bounds = c.getBounds ();
-	cp.setBounds (bounds.x, bounds.y, bounds.width, bounds.height);
-	cp.setVisible (c.isVisible ());
+        Rectangle bounds = c.getBounds ();
+        cp.setBounds (bounds.x, bounds.y, bounds.width, bounds.height);
+        cp.setVisible (c.isVisible ());
       }
   }
 
   protected ButtonPeer createButton (Button b)
   {
+    checkHeadless();
     return new GtkButtonPeer (b);
   }
 
   protected CanvasPeer createCanvas (Canvas c) 
   {
+    checkHeadless();
     return new GtkCanvasPeer (c);
   }
 
   protected CheckboxPeer createCheckbox (Checkbox cb) 
   {
+    checkHeadless();
     return new GtkCheckboxPeer (cb);
   }
 
   protected CheckboxMenuItemPeer createCheckboxMenuItem (CheckboxMenuItem cmi)
   {
+    checkHeadless();
     return new GtkCheckboxMenuItemPeer (cmi);
   }
 
   protected ChoicePeer createChoice (Choice c) 
   {
+    checkHeadless();
     return new GtkChoicePeer (c);
   }
 
   protected DialogPeer createDialog (Dialog d)
   {
+    checkHeadless();
+    GtkMainThread.createWindow();
     return new GtkDialogPeer (d);
   }
 
   protected FileDialogPeer createFileDialog (FileDialog fd)
   {
+    checkHeadless();
     return new GtkFileDialogPeer (fd);
   }
 
   protected FramePeer createFrame (Frame f)
   {
+    checkHeadless();
+    GtkMainThread.createWindow();
     return new GtkFramePeer (f);
   }
 
   protected LabelPeer createLabel (Label label) 
   {
+    checkHeadless();
     return new GtkLabelPeer (label);
   }
 
   protected ListPeer createList (List list)
   {
+    checkHeadless();
     return new GtkListPeer (list);
   }
 
   protected MenuPeer createMenu (Menu m) 
   {
+    checkHeadless();
     return new GtkMenuPeer (m);
   }
 
   protected MenuBarPeer createMenuBar (MenuBar mb) 
   {
+    checkHeadless();
     return new GtkMenuBarPeer (mb);
   }
 
   protected MenuItemPeer createMenuItem (MenuItem mi) 
   {
+    checkHeadless();
     return new GtkMenuItemPeer (mi);
   }
 
   protected PanelPeer createPanel (Panel p) 
   {
+    checkHeadless();
     return new GtkPanelPeer (p);
   }
 
   protected PopupMenuPeer createPopupMenu (PopupMenu target) 
   {
+    checkHeadless();
     return new GtkPopupMenuPeer (target);
   }
 
   protected ScrollPanePeer createScrollPane (ScrollPane sp) 
   {
+    checkHeadless();
     return new GtkScrollPanePeer (sp);
   }
 
   protected ScrollbarPeer createScrollbar (Scrollbar sb) 
   {
+    checkHeadless();
     return new GtkScrollbarPeer (sb);
   }
 
   protected TextAreaPeer createTextArea (TextArea ta) 
   {
+    checkHeadless();
     return new GtkTextAreaPeer (ta);
   }
 
   protected TextFieldPeer createTextField (TextField tf) 
   {
+    checkHeadless();
     return new GtkTextFieldPeer (tf);
   }
 
   protected WindowPeer createWindow (Window w)
   {
+    checkHeadless();
+    GtkMainThread.createWindow();
     return new GtkWindowPeer (w);
   }
 
   public EmbeddedWindowPeer createEmbeddedWindow (EmbeddedWindow w)
   {
+    checkHeadless();
+    GtkMainThread.createWindow();
     return new GtkEmbeddedWindowPeer (w);
   }
 
@@ -565,7 +603,7 @@ public class GtkToolkit extends gnu.java.awt.ClasspathToolkit
    */
   private FontPeer getFontPeer (String name, int style, int size) 
   {
-    Map attrs = new HashMap ();
+    Map<TextAttribute,Object> attrs = new HashMap<TextAttribute,Object>();
     ClasspathFontPeer.copyStyleToAttrs (style, attrs);
     ClasspathFontPeer.copySizeToAttrs (size, attrs);
     return getClasspathFontPeer (name, attrs);
@@ -577,16 +615,17 @@ public class GtkToolkit extends gnu.java.awt.ClasspathToolkit
    * model, hence "ClasspathFontPeer". 
    */
 
-  public ClasspathFontPeer getClasspathFontPeer (String name, Map attrs)
+  public ClasspathFontPeer getClasspathFontPeer (String name,
+                                                 Map<?,?> attrs)
   {
-    Map keyMap = new HashMap (attrs);
+    Map<Object,Object> keyMap = new HashMap<Object,Object>(attrs);
     // We don't know what kind of "name" the user requested (logical, face,
     // family), and we don't actually *need* to know here. The worst case
     // involves failure to consolidate fonts with the same backend in our
     // cache. This is harmless.
     keyMap.put ("GtkToolkit.RequestedFontName", name);
     if (fontCache.containsKey (keyMap))
-      return (ClasspathFontPeer) fontCache.get (keyMap);
+      return fontCache.get (keyMap);
     else
       {
         ClasspathFontPeer newPeer = new GdkFontPeer (name, attrs);
@@ -595,12 +634,6 @@ public class GtkToolkit extends gnu.java.awt.ClasspathToolkit
       }
   }
 
-  public ClasspathTextLayoutPeer getClasspathTextLayoutPeer (AttributedString str, 
-                                                             FontRenderContext frc)
-  {
-    return new GdkTextLayout(str, frc);
-  }
-
   protected EventQueue getSystemEventQueueImpl() 
   {
     synchronized (GtkToolkit.class)
@@ -608,20 +641,45 @@ public class GtkToolkit extends gnu.java.awt.ClasspathToolkit
         if (q == null)
           {
             q = new EventQueue();
-            GtkGenericPeer.enableQueue (q);
           }
       }    
     return q;
   }
 
+  public Cursor createCustomCursor(Image image, Point hotspot, String name)
+  {
+    return new GtkCursor(image, hotspot, name);
+  }
+
   protected native void loadSystemColors (int[] systemColors);
 
   public DragSourceContextPeer createDragSourceContextPeer(DragGestureEvent e)
   {
-    throw new Error("not implemented");
+    if (GraphicsEnvironment.isHeadless())
+      throw new InvalidDnDOperationException();
+    return new GtkDragSourceContextPeer(e);
+  }
+  
+  public <T extends DragGestureRecognizer> T
+  createDragGestureRecognizer(Class<T> recognizer, DragSource ds, 
+                              Component comp, int actions,
+                              DragGestureListener l)
+  {
+    if (recognizer.getName().equals("java.awt.dnd.MouseDragGestureRecognizer")
+        && ! GraphicsEnvironment.isHeadless())
+      {
+        GtkMouseDragGestureRecognizer gestureRecognizer
+          = new GtkMouseDragGestureRecognizer(ds, comp, actions, l);
+        gestureRecognizer.registerListeners();
+        return recognizer.cast(gestureRecognizer);
+      }
+    else
+      {
+        return null;
+      }
   }
 
-  public Map mapInputMethodHighlight(InputMethodHighlight highlight)
+  public Map<TextAttribute,?> mapInputMethodHighlight(InputMethodHighlight highlight)
   {
     throw new Error("not implemented");
   }
@@ -650,10 +708,59 @@ public class GtkToolkit extends gnu.java.awt.ClasspathToolkit
     return new GdkRobotPeer (screen);
   }
 
+  public boolean getLockingKeyState(int keyCode)
+  {
+    int state = getLockState(keyCode);
+    
+    if (state != -1)
+      return state == 1;
+    
+    if (AWTUtilities.isValidKey(keyCode))
+      throw new UnsupportedOperationException
+        ("cannot get locking state of key code " + keyCode);
+    
+    throw new IllegalArgumentException("invalid key code " + keyCode);
+  }
+
+  protected native int getLockState(int keyCode);
+
   public void registerImageIOSpis(IIORegistry reg)
   {
     GdkPixbufDecoder.registerSpis(reg);
   }
 
-  public static native void gtkMain();
+  protected MouseInfoPeer getMouseInfoPeer()
+  {
+    return new GtkMouseInfoPeer();
+  }
+
+  public boolean isFrameStateSupported(int state)
+  {
+    // GTK supports ICONFIED, NORMAL and MAXIMIZE_BOTH, but
+    // not (yet?) MAXIMIZE_VERT and MAXIMIZE_HORIZ.
+    return state == Frame.NORMAL || state == Frame.ICONIFIED
+           || state == Frame.MAXIMIZED_BOTH;
+  }
+
+  private void checkHeadless()
+  {
+    if (GraphicsEnvironment.isHeadless())
+      throw new HeadlessException();
+  }
+
+  public native int getMouseNumberOfButtons();
+
+  @Override
+  public boolean isModalExclusionTypeSupported
+  (Dialog.ModalExclusionType modalExclusionType)
+  {
+    return false;
+  }
+
+  @Override
+  public boolean isModalityTypeSupported(Dialog.ModalityType modalityType)
+  {
+    return false;
+  }
+
 } // class GtkToolkit
